@@ -44,9 +44,11 @@
 
 /* --- structures --- */
 typedef struct {
+  GtkPathType   type;
   GPatternSpec *pspec;
-  gpointer user_data;
-  guint seq_id;
+  GSList       *path;
+  gpointer      user_data;
+  guint         seq_id;
 } PatternSpec;
 
 
@@ -59,17 +61,26 @@ static GQuark		 key_id_class_binding_set = 0;
 
 
 /* --- functions --- */
+static void
+pattern_spec_free (PatternSpec *pspec)
+{
+  _gtk_rc_free_widget_class_path (pspec->path);
+  if (pspec->pspec)
+    g_pattern_spec_free (pspec->pspec);
+  g_free (pspec);
+}
+
 static GtkBindingSignal*
 binding_signal_new (const gchar *signal_name,
 		    guint	 n_args)
 {
   GtkBindingSignal *signal;
-  
-  signal = g_new (GtkBindingSignal, 1);
+
+  signal = (GtkBindingSignal *) g_slice_alloc0 (sizeof (GtkBindingSignal) + n_args * sizeof (GtkBindingArg));
   signal->next = NULL;
-  signal->signal_name = g_strdup (signal_name);
+  signal->signal_name = (gchar *)g_intern_string (signal_name);
   signal->n_args = n_args;
-  signal->args = g_new0 (GtkBindingArg, n_args);
+  signal->args = (GtkBindingArg *)(signal + 1);
   
   return signal;
 }
@@ -84,9 +95,7 @@ binding_signal_free (GtkBindingSignal *sig)
       if (G_TYPE_FUNDAMENTAL (sig->args[i].arg_type) == G_TYPE_STRING)
 	g_free (sig->args[i].d.string_data);
     }
-  g_free (sig->args);
-  g_free (sig->signal_name);
-  g_free (sig);
+  g_slice_free1 (sizeof (GtkBindingSignal) + sig->n_args * sizeof (GtkBindingArg), sig);
 }
 
 static guint
@@ -197,6 +206,7 @@ binding_entry_new (GtkBindingSet  *binding_set,
   entry->binding_set = binding_set,
   entry->destroyed = FALSE;
   entry->in_emission = FALSE;
+  entry->marks_unbound = FALSE;
   entry->signals = NULL;
 
   entry->set_next = binding_set->entries;
@@ -551,15 +561,24 @@ gtk_binding_entry_activate (GtkBindingEntry *entry,
   return handled;
 }
 
+/**
+ * gtk_binding_set_new:
+ * @set_name: unique name of this binding set
+ *
+ * Gtk+ maintains a global list of binding sets. Each binding set has
+ * a unique name which needs to be specified upon creation.
+ *
+ * Return value: new binding set
+ **/
 GtkBindingSet*
-gtk_binding_set_new (const gchar    *set_name)
+gtk_binding_set_new (const gchar *set_name)
 {
   GtkBindingSet *binding_set;
   
   g_return_val_if_fail (set_name != NULL, NULL);
   
   binding_set = g_new (GtkBindingSet, 1);
-  binding_set->set_name = g_strdup (set_name);
+  binding_set->set_name = (gchar *) g_intern_string (set_name);
   binding_set->widget_path_pspecs = NULL;
   binding_set->widget_class_pspecs = NULL;
   binding_set->class_branch_pspecs = NULL;
@@ -572,6 +591,16 @@ gtk_binding_set_new (const gchar    *set_name)
   return binding_set;
 }
 
+/**
+ * gtk_binding_set_by_class:
+ * @object_class: a valid #GtkObject class
+ *
+ * This function returns the binding set named after the type name of
+ * the passed in class structure. New binding sets are created on
+ * demand by this function.
+ *
+ * Return value: the binding set corresponding to @object_class
+ **/
 GtkBindingSet*
 gtk_binding_set_by_class (gpointer object_class)
 {
@@ -598,8 +627,18 @@ gtk_binding_set_by_class (gpointer object_class)
   return binding_set;
 }
 
+/**
+ * gtk_binding_set_find:
+ * @set_name: unique binding set name
+ *
+ * Find a binding set by its globally unique name. The @set_name can
+ * either be a name used for gtk_binding_set_new() or the type name of
+ * a class used in gtk_binding_set_by_class().
+ *
+ * Return value: %NULL or the specified binding set
+ **/
 GtkBindingSet*
-gtk_binding_set_find (const gchar    *set_name)
+gtk_binding_set_find (const gchar *set_name)
 {
   GSList *slist;
   
@@ -616,6 +655,18 @@ gtk_binding_set_find (const gchar    *set_name)
   return NULL;
 }
 
+/**
+ * gtk_binding_set_activate:
+ * @binding_set: @binding_set to activate
+ * @keyval:      key value of the binding
+ * @modifiers:   key modifier of the binding
+ * @object:      object to activate when binding found
+ *
+ * Find a key binding matching @keyval and @modifiers within
+ * @binding_set and activate the binding on @object.
+ *
+ * Return value: %TRUE if a binding was found and activated
+ **/
 gboolean
 gtk_binding_set_activate (GtkBindingSet	 *binding_set,
 			  guint		  keyval,
@@ -637,6 +688,14 @@ gtk_binding_set_activate (GtkBindingSet	 *binding_set,
   return FALSE;
 }
 
+/**
+ * gtk_binding_entry_clear:
+ * @binding_set:
+ * @keyval:
+ * @modifiers:
+ *
+ * Use of this function is deprecated.
+ **/
 void
 gtk_binding_entry_clear (GtkBindingSet	*binding_set,
 			 guint		 keyval,
@@ -656,6 +715,47 @@ gtk_binding_entry_clear (GtkBindingSet	*binding_set,
   entry = binding_entry_new (binding_set, keyval, modifiers);
 }
 
+/**
+ * gtk_binding_entry_skip:
+ * @binding_set: @binding_set to skip an entry of
+ * @keyval:      key value of binding to skip
+ * @modifiers:   key modifier of binding to skip
+ *
+ * Since: 2.12
+ *
+ * Install a binding on @binding_set which causes key lookups
+ * to be aborted, to prevent bindings from lower priority sets
+ * to be activated.
+ **/
+void
+gtk_binding_entry_skip (GtkBindingSet  *binding_set,
+                        guint           keyval,
+                        GdkModifierType modifiers)
+{
+  GtkBindingEntry *entry;
+
+  g_return_if_fail (binding_set != NULL);
+
+  keyval = gdk_keyval_to_lower (keyval);
+  modifiers = modifiers & BINDING_MOD_MASK ();
+
+  entry = binding_ht_lookup_entry (binding_set, keyval, modifiers);
+  if (entry)
+    binding_entry_destroy (entry);
+
+  entry = binding_entry_new (binding_set, keyval, modifiers);
+  entry->marks_unbound = TRUE;
+}
+
+/**
+ * gtk_binding_entry_remove:
+ * @binding_set: @binding_set to remove an entry of
+ * @keyval:      key value of binding to remove
+ * @modifiers:   key modifier of binding to remove
+ *
+ * Remove a binding previously installed via
+ * gtk_binding_entry_add_signal() on @binding_set.
+ **/
 void
 gtk_binding_entry_remove (GtkBindingSet	 *binding_set,
 			  guint		  keyval,
@@ -673,12 +773,34 @@ gtk_binding_entry_remove (GtkBindingSet	 *binding_set,
     binding_entry_destroy (entry);
 }
 
+/**
+ * gtk_binding_entry_add_signall:
+ * @binding_set:  binding set to add a signal to
+ * @keyval:       key value
+ * @modifiers:    key modifier
+ * @signal_name:  signal name to be bound
+ * @binding_args: list of #GtkBindingArg signal arguments
+ *
+ * Deprecated.
+ **/
 void
 gtk_binding_entry_add_signall (GtkBindingSet  *binding_set,
-			       guint	       keyval,
-			       GdkModifierType modifiers,
-			       const gchar    *signal_name,
-			       GSList	      *binding_args)
+                               guint	       keyval,
+                               GdkModifierType modifiers,
+                               const gchar    *signal_name,
+                               GSList	      *binding_args)
+{
+  _gtk_binding_entry_add_signall (binding_set,
+                                  keyval, modifiers,
+                                  signal_name, binding_args);
+}
+
+void
+_gtk_binding_entry_add_signall (GtkBindingSet  *binding_set,
+                                guint	       keyval,
+                                GdkModifierType modifiers,
+                                const gchar    *signal_name,
+                                GSList	      *binding_args)
 {
   GtkBindingEntry *entry;
   GtkBindingSignal *signal, **signal_p;
@@ -751,6 +873,20 @@ gtk_binding_entry_add_signall (GtkBindingSet  *binding_set,
   *signal_p = signal;
 }
 
+/**
+ * gtk_binding_entry_add_signal:
+ * @binding_set: @binding_set to install an entry for
+ * @keyval:      key value of binding to install
+ * @modifiers:   key modifier of binding to install
+ * @signal_name: signal to execute upon activation
+ * @n_args:      number of arguments to @signal_name
+ * @:            arguments to @signal_name
+ *
+ * Override or install a new key binding for @keyval with @modifiers on
+ * @binding_set.  When the binding is activated, @signal_name will be
+ * emitted on the target widget, with @n_args @Varargs used as
+ * arguments.
+ **/
 void
 gtk_binding_entry_add_signal (GtkBindingSet  *binding_set,
 			      guint           keyval,
@@ -772,15 +908,12 @@ gtk_binding_entry_add_signal (GtkBindingSet  *binding_set,
     {
       GtkBindingArg *arg;
 
-      arg = g_new0 (GtkBindingArg, 1);
+      arg = g_slice_new0 (GtkBindingArg);
       slist = g_slist_prepend (slist, arg);
 
       arg->arg_type = va_arg (args, GtkType);
       switch (G_TYPE_FUNDAMENTAL (arg->arg_type))
 	{
-	  /* for elaborated commenting about var args collection, take a look
-	   * at gtk_arg_collect_value() in gtkargcollector.c
-	   */
 	case G_TYPE_CHAR:
 	case G_TYPE_UCHAR:
 	case G_TYPE_INT:
@@ -831,12 +964,22 @@ gtk_binding_entry_add_signal (GtkBindingSet  *binding_set,
   free_slist = slist;
   while (slist)
     {
-      g_free (slist->data);
+      g_slice_free (GtkBindingArg, slist->data);
       slist = slist->next;
     }
   g_slist_free (free_slist);
 }
 
+/**
+ * gtk_binding_set_add_path:
+ * @binding_set:  binding set to add a path to
+ * @path_type:    path type the pattern applies to
+ * @path_pattern: the actual match pattern
+ * @priority:     binding priority
+ *
+ * This function is used internally by the GtkRC parsing mechanism to
+ * assign match patterns to #GtkBindingSet structures.
+ **/
 void
 gtk_binding_set_add_path (GtkBindingSet	     *binding_set,
 			  GtkPathType	      path_type,
@@ -871,7 +1014,18 @@ gtk_binding_set_add_path (GtkBindingSet	     *binding_set,
     }
   
   pspec = g_new (PatternSpec, 1);
-  pspec->pspec = g_pattern_spec_new (path_pattern);
+  pspec->type = path_type;
+  if (path_type == GTK_PATH_WIDGET_CLASS)
+    {
+      pspec->pspec = NULL;
+      pspec->path = _gtk_rc_parse_widget_class_path (path_pattern);
+    }
+  else
+    {
+      pspec->pspec = g_pattern_spec_new (path_pattern);
+      pspec->path = NULL;
+    }
+    
   pspec->seq_id = priority << 28;
   pspec->user_data = binding_set;
   
@@ -887,8 +1041,7 @@ gtk_binding_set_add_path (GtkBindingSet	     *binding_set,
 	{
 	  GtkPathPriorityType lprio = tmp_pspec->seq_id >> 28;
 
-	  g_pattern_spec_free (pspec->pspec);
-	  g_free (pspec);
+	  pattern_spec_free (pspec);
 	  pspec = NULL;
 	  if (lprio < priority)
 	    {
@@ -909,25 +1062,44 @@ static gboolean
 binding_match_activate (GSList          *pspec_list,
 			GtkObject	*object,
 			guint	         path_length,
-			const gchar     *path,
-			const gchar     *path_reversed)
+			gchar           *path,
+			gchar           *path_reversed,
+                        gboolean        *unbound)
 {
   GSList *slist;
 
+  *unbound = FALSE;
+
   for (slist = pspec_list; slist; slist = slist->next)
     {
       PatternSpec *pspec;
+      GtkBindingSet *binding_set;
 
+      binding_set = NULL;
       pspec = slist->data;
-      if (g_pattern_match (pspec->pspec, path_length, path, path_reversed))
-	{
-	  GtkBindingSet *binding_set;
+      
+      if (pspec->type == GTK_PATH_WIDGET_CLASS)
+        {
+          if (_gtk_rc_match_widget_class (pspec->path, path_length, path, path_reversed))
+	    binding_set = pspec->user_data;
+        }
+      else
+        {
+          if (g_pattern_match (pspec->pspec, path_length, path, path_reversed))
+	    binding_set = pspec->user_data;
+        }
 
-	  binding_set = pspec->user_data;
+      if (binding_set)
+        {
+          if (binding_set->current->marks_unbound)
+            {
+              *unbound = TRUE;
+              return FALSE;
+            }
 
-	  if (gtk_binding_entry_activate (binding_set->current, object))
-	    return TRUE;
-	}
+          if (gtk_binding_entry_activate (binding_set->current, object))
+            return TRUE;
+        }
     }
 
   return FALSE;
@@ -1021,13 +1193,17 @@ gtk_bindings_activate_list (GtkObject *object,
       guint path_length;
       gchar *path, *path_reversed;
       GSList *patterns;
+      gboolean unbound;
 
       gtk_widget_path (widget, &path_length, &path, &path_reversed);
       patterns = gtk_binding_entries_sort_patterns (entries, GTK_PATH_WIDGET, is_release);
-      handled = binding_match_activate (patterns, object, path_length, path, path_reversed);
+      handled = binding_match_activate (patterns, object, path_length, path, path_reversed, &unbound);
       g_slist_free (patterns);
       g_free (path);
       g_free (path_reversed);
+
+      if (unbound)
+        return FALSE;
     }
 
   if (!handled)
@@ -1035,43 +1211,66 @@ gtk_bindings_activate_list (GtkObject *object,
       guint path_length;
       gchar *path, *path_reversed;
       GSList *patterns;
+      gboolean unbound;
 
       gtk_widget_class_path (widget, &path_length, &path, &path_reversed);
       patterns = gtk_binding_entries_sort_patterns (entries, GTK_PATH_WIDGET_CLASS, is_release);
-      handled = binding_match_activate (patterns, object, path_length, path, path_reversed);
+      handled = binding_match_activate (patterns, object, path_length, path, path_reversed, &unbound);
       g_slist_free (patterns);
       g_free (path);
       g_free (path_reversed);
+
+      if (unbound)
+        return FALSE;
     }
 
   if (!handled)
     {
       GSList *patterns;
       GType class_type;
-      
+      gboolean unbound = FALSE;
+
       patterns = gtk_binding_entries_sort_patterns (entries, GTK_PATH_CLASS, is_release);
       class_type = G_TYPE_FROM_INSTANCE (object);
       while (class_type && !handled)
 	{
 	  guint path_length;
-	  const gchar *path;
+	  gchar *path;
 	  gchar *path_reversed;
-	  
-	  path = g_type_name (class_type);
+
+	  path = g_strdup (g_type_name (class_type));
 	  path_reversed = g_strdup (path);
 	  g_strreverse (path_reversed);
 	  path_length = strlen (path);
-	  handled = binding_match_activate (patterns, object, path_length, path, path_reversed);
+	  handled = binding_match_activate (patterns, object, path_length, path, path_reversed, &unbound);
+	  g_free (path);
 	  g_free (path_reversed);
 
+          if (unbound)
+            break;
+
 	  class_type = g_type_parent (class_type);
 	}
       g_slist_free (patterns);
+
+      if (unbound)
+        return FALSE;
     }
 
   return handled;
 }
 
+/**
+ * gtk_bindings_activate:
+ * @object: object to activate when binding found
+ * @keyval: key value of the binding
+ * @modifiers: key modifier of the binding
+ *
+ * Find a key binding matching @keyval and @modifiers and activate the
+ * binding on @object.
+ *
+ * Return value: %TRUE if a binding was found and activated
+ **/
 gboolean
 gtk_bindings_activate (GtkObject      *object,
 		       guint	       keyval,
@@ -1308,12 +1507,15 @@ gtk_binding_parse_bind (GScanner       *scanner,
 {
   guint keyval = 0;
   GdkModifierType modifiers = 0;
+  gboolean unbind = FALSE;
 
   g_return_val_if_fail (scanner != NULL, G_TOKEN_ERROR);
   
   g_scanner_get_next_token (scanner);
-  if (scanner->token != GTK_RC_TOKEN_BIND)
+  if (scanner->token != GTK_RC_TOKEN_BIND &&
+      scanner->token != GTK_RC_TOKEN_UNBIND)
     return GTK_RC_TOKEN_BIND;
+  unbind = scanner->token == GTK_RC_TOKEN_UNBIND;
   g_scanner_get_next_token (scanner);
   if (scanner->token != G_TOKEN_STRING)
     return G_TOKEN_STRING;
@@ -1322,7 +1524,14 @@ gtk_binding_parse_bind (GScanner       *scanner,
   if (keyval == 0)
     return G_TOKEN_STRING;
 
+  if (unbind)
+    {
+      gtk_binding_entry_skip (binding_set, keyval, modifiers);
+      return G_TOKEN_NONE;
+    }
+
   g_scanner_get_next_token (scanner);
+
   if (scanner->token != '{')
     return '{';
 
@@ -1354,8 +1563,22 @@ gtk_binding_parse_bind (GScanner       *scanner,
   return G_TOKEN_NONE;
 }
 
+/**
+ * gtk_binding_parse_binding:
+ * @scanner: GtkRC scanner
+ *
+ * Deprecated as public API, used only internally.
+ *
+ * Return value: expected token upon errors, %G_TOKEN_NONE on success.
+ **/
 guint
-gtk_binding_parse_binding (GScanner       *scanner)
+gtk_binding_parse_binding (GScanner *scanner)
+{
+  return _gtk_binding_parse_binding (scanner);
+}
+
+guint
+_gtk_binding_parse_binding (GScanner *scanner)
 {
   gchar *name;
   GtkBindingSet *binding_set;
@@ -1393,6 +1616,7 @@ gtk_binding_parse_binding (GScanner       *scanner)
 	  guint expected_token;
 
 	case GTK_RC_TOKEN_BIND:
+	case GTK_RC_TOKEN_UNBIND:
 	  expected_token = gtk_binding_parse_bind (scanner, binding_set);
 	  if (expected_token != G_TOKEN_NONE)
 	    return expected_token;
@@ -1419,8 +1643,7 @@ free_pattern_specs (GSList *pattern_specs)
 
       pspec = slist->data;
 
-      g_pattern_spec_free (pspec->pspec);
-      g_free (pspec);
+      pattern_spec_free (pspec);
     }
 
   g_slist_free (pattern_specs);
@@ -1443,7 +1666,6 @@ binding_set_delete (GtkBindingSet *binding_set)
   free_pattern_specs (binding_set->widget_class_pspecs);
   free_pattern_specs (binding_set->class_branch_pspecs);
 
-  g_free (binding_set->set_name);
   g_free (binding_set);
 }
 
