@@ -24,6 +24,8 @@
  * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
  */
 
+#include <stdlib.h>
+#include <string.h>
 #include <config.h>
 #include "gtkbutton.h"
 #include "gtkdialog.h"
@@ -37,6 +39,7 @@
 #include "gtkintl.h"
 #include "gtkbindings.h"
 #include "gtkprivate.h"
+#include "gtkbuildable.h"
 #include "gtkalias.h"
 
 #define GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), GTK_TYPE_DIALOG, GtkDialogPrivate))
@@ -76,6 +79,22 @@ static void gtk_dialog_close             (GtkDialog        *dialog);
 
 static ResponseData* get_response_data   (GtkWidget        *widget,
 					  gboolean          create);
+static void gtk_dialog_buildable_interface_init     (GtkBuildableIface *iface);
+static GObject * gtk_dialog_buildable_get_internal_child (GtkBuildable *buildable,
+						     GtkBuilder   *builder,
+						     const gchar  *childname);
+static gboolean gtk_dialog_buildable_custom_tag_start (GtkBuildable  *buildable,
+						       GtkBuilder    *builder,
+						       GObject       *child,
+						       const gchar   *tagname,
+						       GMarkupParser *parser,
+						       gpointer      *data);
+static void gtk_dialog_buildable_custom_finished (GtkBuildable *buildable,
+						  GtkBuilder   *builder,
+						  GObject       *child,
+						  const gchar   *tagname,
+						  gpointer      user_data);
+
 
 enum {
   PROP_0,
@@ -90,7 +109,9 @@ enum {
 
 static guint dialog_signals[LAST_SIGNAL];
 
-G_DEFINE_TYPE (GtkDialog, gtk_dialog, GTK_TYPE_WINDOW)
+G_DEFINE_TYPE_WITH_CODE (GtkDialog, gtk_dialog, GTK_TYPE_WINDOW,
+			 G_IMPLEMENT_INTERFACE (GTK_TYPE_BUILDABLE,
+						gtk_dialog_buildable_interface_init))
 
 static void
 gtk_dialog_class_init (GtkDialogClass *class)
@@ -153,7 +174,7 @@ gtk_dialog_class_init (GtkDialogClass *class)
                                                              P_("Spacing between buttons"),
                                                              0,
                                                              G_MAXINT,
-                                                             10,
+                                                             6,
                                                              GTK_PARAM_READABLE));
   
   gtk_widget_class_install_style_property (widget_class,
@@ -235,6 +256,26 @@ gtk_dialog_init (GtkDialog *dialog)
   gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_CENTER_ON_PARENT);
 }
 
+static void
+gtk_dialog_buildable_interface_init (GtkBuildableIface *iface)
+{
+  iface->get_internal_child = gtk_dialog_buildable_get_internal_child;
+  iface->custom_tag_start = gtk_dialog_buildable_custom_tag_start;
+  iface->custom_finished = gtk_dialog_buildable_custom_finished;
+}
+
+static GObject *
+gtk_dialog_buildable_get_internal_child (GtkBuildable *buildable,
+					 GtkBuilder   *builder,
+					 const gchar  *childname)
+{
+    if (strcmp (childname, "vbox") == 0)
+      return G_OBJECT (GTK_DIALOG (buildable)->vbox);
+    else if (strcmp (childname, "action_area") == 0)
+      return G_OBJECT (GTK_DIALOG (buildable)->action_area);
+
+    return NULL;
+}
 
 static void 
 gtk_dialog_set_property (GObject      *object,
@@ -450,15 +491,15 @@ gtk_dialog_new_empty (const gchar     *title,
  * (#GTK_DIALOG_DESTROY_WITH_PARENT). After @flags, button
  * text/response ID pairs should be listed, with a %NULL pointer ending
  * the list. Button text can be either a stock ID such as
- * #GTK_STOCK_OK, or some arbitrary text.  A response ID can be
+ * #GTK_STOCK_OK, or some arbitrary text. A response ID can be
  * any positive number, or one of the values in the #GtkResponseType
  * enumeration. If the user clicks one of these dialog buttons,
- * #GtkDialog will emit the "response" signal with the corresponding
- * response ID. If a #GtkDialog receives the "delete_event" signal, it
- * will emit "response" with a response ID of #GTK_RESPONSE_DELETE_EVENT.
- * However, destroying a dialog does not emit the "response" signal;
- * so be careful relying on "response" when using
- * the #GTK_DIALOG_DESTROY_WITH_PARENT flag. Buttons are from left to right,
+ * #GtkDialog will emit the #GtkDialog::response signal with the corresponding
+ * response ID. If a #GtkDialog receives the #GtkWidget::delete-event signal, 
+ * it will emit ::response with a response ID of #GTK_RESPONSE_DELETE_EVENT.
+ * However, destroying a dialog does not emit the ::response signal;
+ * so be careful relying on ::response when using the 
+ * #GTK_DIALOG_DESTROY_WITH_PARENT flag. Buttons are from left to right,
  * so the first button in the list will be the leftmost button in the dialog.
  *
  * Here's a simple example:
@@ -541,11 +582,11 @@ action_widget_activated (GtkWidget *widget, GtkDialog *dialog)
  * @response_id: response ID for @child
  * 
  * Adds an activatable widget to the action area of a #GtkDialog,
- * connecting a signal handler that will emit the "response" signal on
- * the dialog when the widget is activated.  The widget is appended to
- * the end of the dialog's action area.  If you want to add a
- * non-activatable widget, simply pack it into the
- * <literal>action_area</literal> field of the #GtkDialog struct.
+ * connecting a signal handler that will emit the #GtkDialog::response 
+ * signal on the dialog when the widget is activated. The widget is 
+ * appended to the end of the dialog's action area. If you want to add a
+ * non-activatable widget, simply pack it into the @action_area field 
+ * of the #GtkDialog struct.
  **/
 void
 gtk_dialog_add_action_widget (GtkDialog *dialog,
@@ -598,9 +639,9 @@ gtk_dialog_add_action_widget (GtkDialog *dialog,
  * 
  * Adds a button with the given text (or a stock button, if @button_text is a
  * stock ID) and sets things up so that clicking the button will emit the
- * "response" signal with the given @response_id. The button is appended to the
- * end of the dialog's action area. The button widget is returned, but usually
- * you don't need it.
+ * #GtkDialog::response signal with the given @response_id. The button is 
+ * appended to the end of the dialog's action area. The button widget is 
+ * returned, but usually you don't need it.
  *
  * Return value: the button widget that was added
  **/
@@ -821,10 +862,10 @@ gtk_dialog_get_has_separator (GtkDialog *dialog)
  * @dialog: a #GtkDialog
  * @response_id: response ID 
  * 
- * Emits the "response" signal with the given response ID. Used to
- * indicate that the user has responded to the dialog in some way;
+ * Emits the #GtkDialog::response signal with the given response ID. 
+ * Used to indicate that the user has responded to the dialog in some way;
  * typically either you or gtk_dialog_run() will be monitoring the
- * "response" signal and take appropriate action.
+ * ::response signal and take appropriate action.
  **/
 void
 gtk_dialog_response (GtkDialog *dialog,
@@ -902,22 +943,23 @@ run_destroy_handler (GtkDialog *dialog, gpointer data)
  * @dialog: a #GtkDialog
  * 
  * Blocks in a recursive main loop until the @dialog either emits the
- * response signal, or is destroyed. If the dialog is destroyed during the call
- * to gtk_dialog_run(), gtk_dialog_returns #GTK_RESPONSE_NONE.
- * Otherwise, it returns the response ID from the "response" signal emission.
+ * #GtkDialog::response signal, or is destroyed. If the dialog is 
+ * destroyed during the call to gtk_dialog_run(), gtk_dialog_run() returns 
+ * #GTK_RESPONSE_NONE. Otherwise, it returns the response ID from the 
+ * ::response signal emission.
+ *
  * Before entering the recursive main loop, gtk_dialog_run() calls
  * gtk_widget_show() on the dialog for you. Note that you still
  * need to show any children of the dialog yourself.
  *
- * During gtk_dialog_run(), the default behavior of "delete_event" is
- * disabled; if the dialog receives "delete_event", it will not be
+ * During gtk_dialog_run(), the default behavior of #GtkWidget::delete-event 
+ * is disabled; if the dialog receives ::delete_event, it will not be
  * destroyed as windows usually are, and gtk_dialog_run() will return
- * #GTK_RESPONSE_DELETE_EVENT. Also, during gtk_dialog_run() the dialog will be
- * modal. You can force gtk_dialog_run() to return at any time by
- * calling gtk_dialog_response() to emit the "response"
- * signal. Destroying the dialog during gtk_dialog_run() is a very bad
- * idea, because your post-run code won't know whether the dialog was
- * destroyed or not.
+ * #GTK_RESPONSE_DELETE_EVENT. Also, during gtk_dialog_run() the dialog 
+ * will be modal. You can force gtk_dialog_run() to return at any time by
+ * calling gtk_dialog_response() to emit the ::response signal. Destroying 
+ * the dialog during gtk_dialog_run() is a very bad idea, because your 
+ * post-run code won't know whether the dialog was destroyed or not.
  *
  * After gtk_dialog_run() returns, you are responsible for hiding or
  * destroying the dialog if you wish to do so.
@@ -980,7 +1022,7 @@ gtk_dialog_run (GtkDialog *dialog)
   
   delete_handler =
     g_signal_connect (dialog,
-                      "delete_event",
+                      "delete-event",
                       G_CALLBACK (run_delete_handler),
                       &ri);
   
@@ -1115,9 +1157,10 @@ gtk_dialog_set_alternative_button_order_valist (GtkDialog *dialog,
  * @first_response_id: a response id used by one @dialog's buttons
  * @Varargs: a list of more response ids of @dialog's buttons, terminated by -1
  *
- * Sets an alternative button order. If the gtk-alternative-button-order 
- * setting is set to %TRUE, the dialog buttons are reordered according to 
- * the order of the response ids passed to this function.
+ * Sets an alternative button order. If the 
+ * #GtkSettings:gtk-alternative-button-order setting is set to %TRUE, 
+ * the dialog buttons are reordered according to the order of the 
+ * response ids passed to this function.
  *
  * By default, GTK+ dialogs use the button order advocated by the Gnome 
  * <ulink url="http://developer.gnome.org/projects/gup/hig/2.0/">Human 
@@ -1179,9 +1222,10 @@ gtk_dialog_set_alternative_button_order (GtkDialog *dialog,
  * @n_params: the number of response ids in @new_order
  * @new_order: an array of response ids of @dialog's buttons
  *
- * Sets an alternative button order. If the gtk-alternative-button-order 
- * setting is set to %TRUE, the dialog buttons are reordered according to 
- * the order of the response ids in @new_order.
+ * Sets an alternative button order. If the 
+ * #GtkSettings:gtk-alternative-button-order setting is set to %TRUE, 
+ * the dialog buttons are reordered according to the order of the 
+ * response ids in @new_order.
  *
  * See gtk_dialog_set_alternative_button_order() for more information.
  *
@@ -1213,5 +1257,134 @@ gtk_dialog_set_alternative_button_order_from_array (GtkDialog *dialog,
     }
 }
 
+typedef struct {
+  gchar *widget_name;
+  gchar *response_id;
+} ActionWidgetInfo;
+
+typedef struct {
+  GtkDialog *dialog;
+  GtkBuilder *builder;
+  GSList *items;
+  gchar *response;
+} ActionWidgetsSubParserData;
+
+static void
+attributes_start_element (GMarkupParseContext *context,
+			  const gchar         *element_name,
+			  const gchar        **names,
+			  const gchar        **values,
+			  gpointer             user_data,
+			  GError             **error)
+{
+  ActionWidgetsSubParserData *parser_data = (ActionWidgetsSubParserData*)user_data;
+  guint i;
+
+  if (strcmp (element_name, "action-widget") == 0)
+    for (i = 0; names[i]; i++)
+      if (strcmp (names[i], "response") == 0)
+	parser_data->response = g_strdup (values[i]);
+  else if (strcmp (element_name, "action-widgets") == 0)
+    return;
+  else
+    g_warning ("Unsupported tag for GtkDialog: %s\n", element_name);
+}
+
+static void
+attributes_text_element (GMarkupParseContext *context,
+			 const gchar         *text,
+			 gsize                text_len,
+			 gpointer             user_data,
+			 GError             **error)
+{
+  ActionWidgetsSubParserData *parser_data = (ActionWidgetsSubParserData*)user_data;
+  ActionWidgetInfo *item;
+
+  if (!parser_data->response)
+    return;
+
+  item = g_new (ActionWidgetInfo, 1);
+  item->widget_name = g_strndup (text, text_len);
+  item->response_id = parser_data->response;
+  parser_data->items = g_slist_prepend (parser_data->items, item);
+  parser_data->response = NULL;
+}
+
+static const GMarkupParser attributes_parser =
+  {
+    attributes_start_element,
+    NULL,
+    attributes_text_element,
+  };
+
+gboolean
+gtk_dialog_buildable_custom_tag_start (GtkBuildable  *buildable,
+				       GtkBuilder    *builder,
+				       GObject       *child,
+				       const gchar   *tagname,
+				       GMarkupParser *parser,
+				       gpointer      *data)
+{
+  ActionWidgetsSubParserData *parser_data;
+
+  if (child)
+    return FALSE;
+
+  if (strcmp (tagname, "action-widgets") == 0)
+    {
+      parser_data = g_slice_new0 (ActionWidgetsSubParserData);
+      parser_data->dialog = GTK_DIALOG (buildable);
+      parser_data->items = NULL;
+
+      *parser = attributes_parser;
+      *data = parser_data;
+      return TRUE;
+    }
+
+  return FALSE;
+}
+
+static void
+gtk_dialog_buildable_custom_finished (GtkBuildable *buildable,
+				      GtkBuilder   *builder,
+				      GObject      *child,
+				      const gchar  *tagname,
+				      gpointer      user_data)
+{
+  GSList *l;
+  ActionWidgetsSubParserData *parser_data;
+  GObject *object;
+  ResponseData* data;
+
+  if (strcmp (tagname, "action-widgets"))
+    return;
+
+  parser_data = (ActionWidgetsSubParserData*)user_data;
+  parser_data->items = g_slist_reverse (parser_data->items);
+
+  for (l = parser_data->items; l; l = l->next)
+    {
+      ActionWidgetInfo *item = l->data;
+
+      object = gtk_builder_get_object (builder, item->widget_name);
+      if (!object)
+	{
+	  g_warning ("Unknown object %s specified in action-widgets %s",
+		     item->widget_name,
+		     gtk_buildable_get_name (GTK_BUILDABLE (object)));
+	  continue;
+	}
+
+      data = get_response_data (GTK_WIDGET (object), TRUE);
+      data->response_id = atoi (item->response_id);
+
+      g_free (item->widget_name);
+      g_free (item->response_id);
+      g_free (item);
+    }
+  g_slist_free (parser_data->items);
+  g_slice_free (ActionWidgetsSubParserData, parser_data);
+}
+
 #define __GTK_DIALOG_C__
 #include "gtkaliasdef.c"
