@@ -104,10 +104,10 @@ struct _GtkToolbar
   /*< private >*/
   gint             button_maxw;		/* maximum width of homogeneous children */
   gint             button_maxh;		/* maximum height of homogeneous children */
-  
-  guint            style_set_connection;
-  guint            icon_size_connection;
-  
+
+  guint            _gtk_reserved1;
+  guint            _gtk_reserved2;
+
   guint            style_set : 1;
   guint            icon_size_set : 1;
 };
@@ -162,7 +162,7 @@ gint            gtk_toolbar_get_drop_index          (GtkToolbar      *toolbar,
 						     gint             y);
 void            gtk_toolbar_set_drop_highlight_item (GtkToolbar      *toolbar,
 						     GtkToolItem     *tool_item,
-						     gint             index);
+						     gint             index_);
 
 /* internal functions */
 gchar *         _gtk_toolbar_elide_underscores      (const gchar     *original);
@@ -171,15 +171,14 @@ void            _gtk_toolbar_paint_space_line       (GtkWidget       *widget,
 						     GdkRectangle    *area,
 						     GtkAllocation   *allocation);
 gint            _gtk_toolbar_get_default_space_size (void);
+void            _gtk_toolbar_rebuild_menu           (GtkToolbar      *toolbar);
 
 
-
-/* deprecated */
-#ifndef GTK_DISABLE_DEPRECATED
 void       gtk_toolbar_set_icon_size   (GtkToolbar      *toolbar,
 				        GtkIconSize      icon_size);
 void       gtk_toolbar_unset_icon_size (GtkToolbar      *toolbar);
 
+#ifndef GTK_DISABLE_DEPRECATED
 /* Simple button items */
 GtkWidget* gtk_toolbar_append_item   (GtkToolbar      *toolbar,
 				      const char      *text,
